import os
import shutil
from typing import Any, Dict, Tuple, Union

from flask import Blueprint, current_app, request, send_file
from sqlalchemy import asc, desc, exc

from fittrackee import db, limiter
from fittrackee.emails.tasks import (
    email_updated_to_new_address,
    password_change_email,
    reset_password_email,
)
from fittrackee.equipments.models import Equipment
from fittrackee.files import get_absolute_file_path
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    UserNotFoundErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.utils import get_readable_duration
from fittrackee.workouts.models import Record, Workout, WorkoutSegment

from .exceptions import InvalidEmailException, UserNotFoundException
from .models import User, UserDataExport, UserSportPreference
from .utils.admin import UserManagerService
from .utils.language import get_language

users_blueprint = Blueprint('users', __name__)

USER_PER_PAGE = 10


@users_blueprint.route('/users', methods=['GET'])
@require_auth(scopes=['users:read'], as_admin=True)
def get_users(auth_user: User) -> Dict:
    """
    Get all users (regardless their account status), if authenticated user
    has admin rights.

    It returns user preferences only for authenticated user.

    **Scope**: ``users:read``

    **Example request**:

    - without parameters:

    .. sourcecode:: http

      GET /api/users HTTP/1.1
      Content-Type: application/json

    - with some query parameters:

    .. sourcecode:: http

      GET /api/users?order_by=workouts_count&par_page=5  HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "users": [
            {
              "admin": true,
              "bio": null,
              "birth_date": null,
              "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
              "email": "admin@example.com",
              "first_name": null,
              "is_admin": true,
              "imperial_units": false,
              "language": "en",
              "last_name": null,
              "location": null,
              "nb_sports": 3,
              "nb_workouts": 6,
              "picture": false,
              "records": [
                {
                  "id": 9,
                  "record_type": "AS",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 10,
                  "record_type": "FD",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 13,
                  "record_type": "HA",
                  "sport_id": 1,
                  "user": "Sam",
                  "value": 43.97,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 11,
                  "record_type": "LD",
                  "sport_id": 1,
                  "user": "admin",
                  "value": "1:01:00",
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                },
                {
                  "id": 12,
                  "record_type": "MS",
                  "sport_id": 1,
                  "user": "admin",
                  "value": 18,
                  "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                  "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
                }
              ],
              "sports_list": [
                  1,
                  4,
                  6
              ],
              "timezone": "Europe/Paris",
              "total_distance": 67.895,
              "total_duration": "6:50:27",
              "username": "admin",
              "weekm": false
            },
            {
              "admin": false,
              "bio": null,
              "birth_date": null,
              "created_at": "Sat, 20 Jul 2019 11:27:03 GMT",
              "email": "sam@example.com",
              "first_name": null,
              "is_admin": false,
              "language": "fr",
              "last_name": null,
              "location": null,
              "nb_sports": 0,
              "nb_workouts": 0,
              "picture": false,
              "records": [],
              "sports_list": [],
              "timezone": "Europe/Paris",
              "total_distance": 0,
              "total_duration": "0:00:00",
              "username": "sam"
            }
          ]
        },
        "status": "success"
      }

    :query integer page: page if using pagination (default: 1)
    :query integer per_page: number of users per page (default: 10, max: 50)
    :query string q: query on user name
    :query string order: sorting order: ``asc``, ``desc`` (default: ``asc``)
    :query string order_by: sorting criteria: ``username``, ``created_at``,
                            ``workouts_count``, ``admin``, ``is_active``
                            (default: ``username``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``

    """
    params = request.args.copy()
    page = int(params.get('page', 1))
    per_page = int(params.get('per_page', USER_PER_PAGE))
    if per_page > 50:
        per_page = 50
    user_column = getattr(User, params.get('order_by', 'username'))
    order = params.get('order', 'asc')
    query = params.get('q')
    users_pagination = (
        User.query.filter(
            User.username.ilike('%' + query + '%') if query else True,
        )
        .order_by(asc(user_column) if order == 'asc' else desc(user_column))
        .paginate(page=page, per_page=per_page, error_out=False)
    )
    users = users_pagination.items
    return {
        'status': 'success',
        'data': {'users': [user.serialize(auth_user) for user in users]},
        'pagination': {
            'has_next': users_pagination.has_next,
            'has_prev': users_pagination.has_prev,
            'page': users_pagination.page,
            'pages': users_pagination.pages,
            'total': users_pagination.total,
        },
    }


@users_blueprint.route('/users/<user_name>', methods=['GET'])
@require_auth(scopes=['users:read'])
def get_single_user(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Get single user details. Only user with admin rights can get other users
    details.

    It returns user preferences only for authenticated user.

    **Scope**: ``users:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/users/admin HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": [
          {
            "admin": true,
            "bio": null,
            "birth_date": null,
            "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
            "email": "admin@example.com",
            "first_name": null,
            "imperial_units": false,
            "is_admin": true,
            "language": "en",
            "last_name": null,
            "location": null,
            "nb_sports": 3,
            "nb_workouts": 6,
            "picture": false,
            "records": [
              {
                "id": 9,
                "record_type": "AS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 10,
                "record_type": "FD",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 13,
                "record_type": "HA",
                "sport_id": 1,
                "user": "Sam",
                "value": 43.97,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 11,
                "record_type": "LD",
                "sport_id": 1,
                "user": "admin",
                "value": "1:01:00",
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 12,
                "record_type": "MS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              }
            ],
            "sports_list": [
                1,
                4,
                6
            ],
            "timezone": "Europe/Paris",
            "total_distance": 67.895,
            "total_duration": "6:50:27",
            "username": "admin"
          }
        ],
        "status": "success"
      }

    :param integer user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404:
        - ``user does not exist``
    """
    if user_name != auth_user.username and not auth_user.admin:
        return ForbiddenErrorResponse()

    try:
        user = User.query.filter_by(username=user_name).first()
        if user:
            return {
                'status': 'success',
                'data': {'users': [user.serialize(auth_user)]},
            }
    except ValueError:
        pass
    return UserNotFoundErrorResponse()


@users_blueprint.route('/users/<user_name>/picture', methods=['GET'])
@limiter.exempt
def get_picture(user_name: str) -> Any:
    """get user picture

    **Example request**:

    .. sourcecode:: http

      GET /api/users/admin/picture HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: image/jpeg

    :param integer user_name: user name

    :statuscode 200: ``success``
    :statuscode 404:
        - ``user does not exist``
        - ``No picture.``

    """
    try:
        user = User.query.filter_by(username=user_name).first()
        if not user:
            return UserNotFoundErrorResponse()
        if user.picture is not None:
            picture_path = get_absolute_file_path(user.picture)
            return send_file(picture_path)
    except Exception:  # nosec
        pass
    return NotFoundErrorResponse('No picture.')


@users_blueprint.route('/users/<user_name>', methods=['PATCH'])
@require_auth(scopes=['users:write'], as_admin=True)
def update_user(auth_user: User, user_name: str) -> Union[Dict, HttpResponse]:
    """
    Update user account.

    - add/remove admin rights (regardless user account status)
    - reset password (and send email to update user password,
      if sending enabled)
    - update user email (and send email to new user email, if sending enabled)
    - activate account for an inactive user

    Only user with admin rights can modify another user.

    **Scope**: ``users:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/users/<user_name> HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": [
          {
            "admin": true,
            "bio": null,
            "birth_date": null,
            "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
            "email": "admin@example.com",
            "first_name": null,
            "imperial_units": false,
            "is_active": true,
            "language": "en",
            "last_name": null,
            "location": null,
            "nb_workouts": 6,
            "nb_sports": 3,
            "picture": false,
            "records": [
              {
                "id": 9,
                "record_type": "AS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 10,
                "record_type": "FD",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 13,
                "record_type": "HA",
                "sport_id": 1,
                "user": "Sam",
                "value": 43.97,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 11,
                "record_type": "LD",
                "sport_id": 1,
                "user": "admin",
                "value": "1:01:00",
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              },
              {
                "id": 12,
                "record_type": "MS",
                "sport_id": 1,
                "user": "admin",
                "value": 18,
                "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
                "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
              }
            ],
            "sports_list": [
                1,
                4,
                6
            ],
            "timezone": "Europe/Paris",
            "total_distance": 67.895,
            "total_duration": "6:50:27",
            "username": "admin"
          }
        ],
        "status": "success"
      }

    :param string user_name: user name

    :<json boolean activate: activate user account
    :<json boolean admin: does the user have administrator rights
    :<json boolean new_email: new user email
    :<json boolean reset_password: reset user password

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``valid email must be provided``
        - ``new email must be different than current email``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403: ``you do not have permissions``
    :statuscode 404: ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    user_data = request.get_json()
    if not user_data:
        return InvalidPayloadErrorResponse()

    try:
        reset_password = user_data.get('reset_password', False)
        new_email = user_data.get('new_email')
        user_manager_service = UserManagerService(username=user_name)
        user, _, _ = user_manager_service.update(
            is_admin=user_data.get('admin'),
            activate=user_data.get('activate', False),
            reset_password=reset_password,
            new_email=new_email,
            with_confirmation=current_app.config['CAN_SEND_EMAILS'],
        )

        if current_app.config['CAN_SEND_EMAILS']:
            user_language = get_language(user.language)
            ui_url = current_app.config['UI_URL']
            if reset_password:
                user_data = {
                    'language': user_language,
                    'email': user.email,
                }
                password_change_email.send(
                    user_data,
                    {
                        'username': user.username,
                        'fittrackee_url': ui_url,
                    },
                )
                password_reset_token = user.encode_password_reset_token(
                    user.id
                )
                reset_password_email.send(
                    user_data,
                    {
                        'expiration_delay': get_readable_duration(
                            current_app.config[
                                'PASSWORD_TOKEN_EXPIRATION_SECONDS'
                            ],
                            user_language,
                        ),
                        'username': user.username,
                        'password_reset_url': (
                            f'{ui_url}/password-reset?'
                            f'token={password_reset_token}'
                        ),
                        'fittrackee_url': ui_url,
                    },
                )

            if new_email:
                user_data = {
                    'language': user_language,
                    'email': user.email_to_confirm,
                }
                email_data = {
                    'username': user.username,
                    'fittrackee_url': ui_url,
                    'email_confirmation_url': (
                        f'{ui_url}/email-update'
                        f'?token={user.confirmation_token}'
                    ),
                }
                email_updated_to_new_address.send(user_data, email_data)

        return {
            'status': 'success',
            'data': {'users': [user.serialize(auth_user)]},
        }
    except UserNotFoundException:
        return UserNotFoundErrorResponse()
    except InvalidEmailException as e:
        return InvalidPayloadErrorResponse(str(e))
    except exc.StatementError as e:
        return handle_error_and_return_response(e, db=db)


@users_blueprint.route('/users/<user_name>', methods=['DELETE'])
@require_auth(scopes=['users:write'])
def delete_user(
    auth_user: User, user_name: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete a user account.

    A user can only delete his own account.

    An admin can delete all accounts except his account if he's the only
    one admin.

    **Scope**: ``users:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/users/john_doe HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: user account deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you can not delete your account, no other user has admin rights``
    :statuscode 404: ``user does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    try:
        user = User.query.filter_by(username=user_name).first()
        if not user:
            return UserNotFoundErrorResponse()

        if user.id != auth_user.id and not auth_user.admin:
            return ForbiddenErrorResponse()
        if (
            user.admin is True
            and User.query.filter_by(admin=True).count() == 1
        ):
            return ForbiddenErrorResponse(
                'you can not delete your account, '
                'no other user has admin rights'
            )

        db.session.query(UserSportPreference).filter(
            UserSportPreference.user_id == user.id
        ).delete()
        db.session.query(Record).filter(Record.user_id == user.id).delete()
        # delete all equipment associated with this user
        db.session.query(Equipment).filter(
            Equipment.user_id == user.id
        ).delete()
        db.session.query(WorkoutSegment).filter(
            WorkoutSegment.workout_id == Workout.id, Workout.user_id == user.id
        ).delete(synchronize_session=False)
        db.session.query(Workout).filter(Workout.user_id == user.id).delete()
        db.session.query(UserDataExport).filter(
            UserDataExport.user_id == user.id
        ).delete()
        db.session.flush()
        user_picture = user.picture
        db.session.delete(user)
        db.session.commit()
        if user_picture:
            picture_path = get_absolute_file_path(user.picture)
            if os.path.isfile(picture_path):
                os.remove(picture_path)
        shutil.rmtree(
            get_absolute_file_path(f'exports/{user.id}'),
            ignore_errors=True,
        )
        shutil.rmtree(
            get_absolute_file_path(f'workouts/{user.id}'),
            ignore_errors=True,
        )
        shutil.rmtree(
            get_absolute_file_path(f'pictures/{user.id}'),
            ignore_errors=True,
        )
        return {'status': 'no content'}, 204
    except (
        exc.IntegrityError,
        exc.OperationalError,
        ValueError,
        OSError,
    ) as e:
        return handle_error_and_return_response(e, db=db)



from fittrackee.exceptions import GenericException


class InvalidGPXException(GenericException): ...


class WorkoutException(GenericException): ...


class WorkoutGPXException(GenericException): ...



from datetime import datetime
from typing import Dict
from uuid import uuid4

from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.sql.expression import text

from fittrackee import db
from fittrackee.utils import encode_uuid

BaseModel: DeclarativeMeta = db.Model


WorkoutEquipment = db.Table(
    'workout_equipments',
    db.Column(
        'workout_id',
        db.Integer,
        db.ForeignKey('workouts.id', ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        'equipment_id',
        db.Integer,
        db.ForeignKey('equipments.id', ondelete="CASCADE"),
        primary_key=True,
        index=True,
    ),
)


class Equipment(BaseModel):
    __tablename__ = 'equipments'
    # a single user can only have one equipment with the same label
    __table_args__ = (
        UniqueConstraint(
            'label',
            'user_id',
            name='equipment_user_label_unique',
        ),
    )

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    uuid = db.Column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    user_id = db.Column(
        db.Integer, db.ForeignKey('users.id'), index=True, nullable=False
    )
    label = db.Column(db.String(50), unique=False, nullable=False)
    description = db.Column(db.String(200), default=None, nullable=True)
    equipment_type_id = db.Column(
        db.Integer, db.ForeignKey('equipment_types.id')
    )
    creation_date = db.Column(db.DateTime, default=datetime.utcnow)
    is_active = db.Column(db.Boolean, default=True, nullable=False)
    total_distance = db.Column(
        db.Numeric(10, 3),
        nullable=True,
        server_default=text('0.0'),  # kilometers
    )
    total_duration = db.Column(
        db.Interval, nullable=False, server_default=text("'00:00:00'")
    )
    total_moving = db.Column(
        db.Interval, nullable=False, server_default=text("'00:00:00'")
    )
    total_workouts = db.Column(
        db.Integer, nullable=False, server_default=text('0')
    )

    workouts = db.relationship(
        'Workout', secondary=WorkoutEquipment, back_populates='equipments'
    )

    def __repr__(self) -> str:
        return f'<Equipment {self.id!r} {self.label!r}>'

    def __init__(
        self,
        label: str,
        equipment_type_id: int,
        description: str,
        user_id: int,
        is_active: bool,
    ) -> None:
        self.label = label
        self.equipment_type_id = equipment_type_id
        self.description = description
        self.user_id = user_id
        self.is_active = is_active

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self) -> Dict:
        serialized_equipment = {
            'id': self.short_id,
            'user_id': self.user_id,
            'label': self.label,
            'equipment_type': self.equipment_type.serialize(),
            'description': self.description,
            'default_for_sport_ids': [
                sport_preference.sport_id
                for sport_preference in self.default_for_sports
            ],
            'creation_date': self.creation_date,
            'is_active': self.is_active,
            'total_distance': (
                0.0
                if self.total_distance is None
                else float(self.total_distance)
            ),
            'total_duration': (
                '0:00:00'
                if self.total_duration is None
                else str(self.total_duration)
            ),
            'total_moving': (
                '0:00:00'
                if self.total_moving is None
                else str(self.total_moving)
            ),
            'workouts_count': (
                0 if self.total_workouts is None else self.total_workouts
            ),
        }
        return serialized_equipment


class EquipmentType(BaseModel):
    __tablename__ = 'equipment_types'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    label = db.Column(db.String(50), unique=False, nullable=False)
    is_active = db.Column(db.Boolean, default=True, nullable=False)
    equipments = db.relationship(
        'Equipment',
        lazy='select',
        backref=db.backref(
            'equipment_type', lazy='joined', single_parent=True
        ),
    )

    def __repr__(self) -> str:
        return f'<EquipmentType {self.label!r}>'

    def __init__(self, label: str, is_active: bool) -> None:
        self.label = label
        self.is_active = is_active

    def serialize(self, is_admin: bool = False) -> Dict:
        serialized_equipment_type = {
            'id': self.id,
            'label': self.label,
            'is_active': self.is_active,
        }
        if is_admin:
            serialized_equipment_type['has_equipments'] = (
                len(self.equipments) > 0
            )
        return serialized_equipment_type



from typing import Optional

from fittrackee.responses import ForbiddenErrorResponse, HttpResponse


def can_view_workout(
    auth_user_id: int, workout_user_id: int
) -> Optional[HttpResponse]:
    """
    Return error response if user has no right to view workout
    """
    if auth_user_id != workout_user_id:
        return ForbiddenErrorResponse()
    return None



# source for StandaloneApplication class:
# http://docs.gunicorn.org/en/stable/custom.html
import os
from typing import Dict, Optional

import gunicorn.app.base
from flask import Flask

from fittrackee import create_app

HOST = os.getenv('HOST', '127.0.0.1')
PORT = os.getenv('PORT', '5000')
WORKERS = os.getenv('APP_WORKERS', 1)
BASEDIR = os.path.abspath(os.path.dirname(__file__))
app = create_app()


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(
        self, current_app: Flask, options: Optional[Dict] = None
    ) -> None:
        self.options = options or {}
        self.application = current_app
        super().__init__()

    def load_config(self) -> None:
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> Flask:
        return self.application


def main() -> None:
    options = {'bind': f'{HOST}:{PORT}', 'workers': WORKERS}
    StandaloneApplication(app, options).run()


if __name__ == '__main__':
    app.run()



from authlib.integrations.flask_oauth2 import AuthorizationServer
from authlib.integrations.sqla_oauth2 import (
    create_query_client_func,
    create_save_token_func,
)

from fittrackee import db

from .models import OAuth2Client, OAuth2Token
from .resource_protector import CustomResourceProtector

query_client = create_query_client_func(db.session, OAuth2Client)
save_token = create_save_token_func(db.session, OAuth2Token)
authorization_server = AuthorizationServer(
    query_client=query_client,
    save_token=save_token,
)
require_auth = CustomResourceProtector()



"""add 'bounds' column to 'Activity' table

Revision ID: 92adde6ac0d0
Revises: dd73d23a7a3d
Create Date: 2018-05-14 14:25:04.889189

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '92adde6ac0d0'
down_revision = 'dd73d23a7a3d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('activities', sa.Column('bounds', postgresql.ARRAY(sa.Float()), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('activities', 'bounds')
    # ### end Alembic commands ###



from datetime import timedelta
from typing import Optional, Union


def convert_in_duration(value: str) -> timedelta:
    hours = int(value.split(':')[0])
    minutes = int(value.split(':')[1])
    return timedelta(seconds=(hours * 3600 + minutes * 60))


def convert_timedelta_to_integer(value: str) -> int:
    hours, minutes, seconds = str(value).split(':')
    return int(hours) * 3600 + int(minutes) * 60 + int(seconds)


def convert_value_to_integer(
    record_type: str, val: Union[str, float]
) -> Optional[int]:
    if val is None:
        return None

    if record_type == 'LD':
        return convert_timedelta_to_integer(str(val))
    elif record_type in ['AS', 'MS']:
        return int(val * 100)
    else:  # 'FD'
        return int(val * 1000)



from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import exc

from fittrackee import db
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    DataNotFoundErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.models import User

from ..equipments.models import EquipmentType

equipment_types_blueprint = Blueprint('equipment_types', __name__)


@equipment_types_blueprint.route('/equipment-types', methods=['GET'])
@require_auth(scopes=['equipments:read'])
def get_equipment_types(auth_user: User) -> Dict:
    """
    Get all types of equipment

    **Scope**: ``equipments:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/equipment-types HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - for non admin user :

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipment_types": [
            {
              "id": 1,
              "is_active": true,
              "label": "Shoe"
            },
            {
              "id": 2,
              "is_active": true,
              "label": "Bike"
            },
            {
              "id": 3,
              "is_active": true,
              "label": "Bike Trainer"
            },
            {
              "id": 4,
              "is_active": true,
              "label": "Kayak_Boat"
            },
            {
              "id": 5,
              "is_active": true,
              "label": "Skis"
            },
            {
              "id": 5,
              "is_active": true,
              "label": "Snowshoes"
            }
          ]
        },
        "status": "success"
      }

    - for admin user :

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipment_types": [
            {
              "has_equipments": true,
              "id": 1,
              "is_active": true,
              "label": "Shoe"
            },
            {
              "has_equipments": true,
              "id": 2,
              "is_active": true,
              "label": "Bike"
            },
            {
              "has_equipments": false,
              "id": 3,
              "is_active": true,
              "label": "Bike Trainer"
            },
            {
              "has_equipments": false,
              "id": 4,
              "is_active": true,
              "label": "Kayak_Boat"
            },
            {
              "has_equipments": false,
              "id": 5,
              "is_active": true,
              "label": "Skis"
            },
            {
              "has_equipments": false,
              "id": 6,
              "is_active": true,
              "label": "Snowshoes"
            }
          ]
        },
        "status": "success"
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403: ``you do not have permissions``

    """
    equipment_types = (
        EquipmentType.query.filter(
            EquipmentType.is_active == True  # noqa
            if not auth_user.admin
            else True
        )
        .order_by(EquipmentType.id)
        .all()
    )
    equipment_types_data = []
    for equipment_type in equipment_types:
        equipment_types_data.append(
            equipment_type.serialize(
                is_admin=auth_user.admin,
            )
        )
    return {
        'status': 'success',
        'data': {'equipment_types': equipment_types_data},
    }


@equipment_types_blueprint.route(
    '/equipment-types/<int:equipment_type_id>', methods=['GET']
)
@require_auth(scopes=['equipments:read'])
def get_equipment_type(
    auth_user: User, equipment_type_id: int
) -> Union[Dict, HttpResponse]:
    """
    Get a type of equipment

    **Scope**: ``equipments:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/equipment-types/2 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - success for non admin user :

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipment_types": [
            {
              "id": 2,
              "is_active": true,
              "label": "Bike"
            }
          ]
        },
        "status": "success"
      }

    - success for admin user :

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipment_types": [
            {
              "has_equipments": true,
              "id": 2,
              "is_active": true,
              "label": "Bike"
            }
          ]
        },
        "status": "success"
      }

    - equipment type not found

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "status": "not found",
        "data": {
          "equipment_types": []
        }
      }

    :param integer equipment_type_id: equipment type id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403: ``you do not have permissions``
    :statuscode 404: ``equipment_type not found``

    """
    equipment_type = EquipmentType.query.filter_by(
        id=equipment_type_id
    ).first()
    if equipment_type:
        if equipment_type.is_active is False and not auth_user.admin:
            return DataNotFoundErrorResponse('equipment_types')
        return {
            'status': 'success',
            'data': {
                'equipment_types': [
                    equipment_type.serialize(
                        is_admin=auth_user.admin,
                    )
                ]
            },
        }
    return DataNotFoundErrorResponse('equipment_types')


@equipment_types_blueprint.route(
    '/equipment-types/<int:equipment_type_id>', methods=['PATCH']
)
@require_auth(scopes=['equipments:write'], as_admin=True)
def update_equipment_type(
    auth_user: User, equipment_type_id: int
) -> Union[Dict, HttpResponse]:
    """
    Update a type of equipment to (de)activate it.

    Authenticated user must be an admin.

    **Scope**: ``equipments:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/equipment-types/2 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - success :

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "equipment_types": [
            {
              "has_equipments": true,
              "id": 2,
              "is_active": true,
              "label": "Bike"
            }
          ]
        },
        "status": "success"
      }

    :param integer equipment_type_id: equipment type id

    :<json boolean is_active: equipment type active status

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 400: ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403: ``you do not have permissions``
    :statuscode 404: ``equipment_type not found``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    data = request.get_json()
    if not data or data.get('is_active') is None:
        return InvalidPayloadErrorResponse()

    try:
        equipment_type = EquipmentType.query.filter_by(
            id=equipment_type_id
        ).first()
        if not equipment_type:
            return DataNotFoundErrorResponse('equipment_types')
        equipment_type.is_active = data.get('is_active')
        db.session.commit()
        return {
            'status': 'success',
            'data': {
                'equipment_types': [
                    equipment_type.serialize(
                        is_admin=auth_user.admin,
                    )
                ]
            },
        }
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)



"""rename 'activity' with 'workout'

Revision ID: 4e8597c50064
Revises: 3243cd25eca7
Create Date: 2021-01-09 19:41:26.589237

"""
import os

import sqlalchemy as sa
from alembic import op
from flask import current_app
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '4e8597c50064'
down_revision = '3243cd25eca7'
branch_labels = None
depends_on = None


def rename_upload_folder(src, dst):
    upload_folder = current_app.config['UPLOAD_FOLDER']
    src_directory = f'{upload_folder}/{src}'
    if os.path.exists(src_directory):
        try:
            os.rename(src_directory, f'{upload_folder}/{dst}')
        except Exception as e:
            print(
                f'ERROR: can not rename upload folder \'{src}\' to \'{dst}\':'
                f'\n      {e}.'
                f'\n      Please rename it manually.')


def upgrade():

    op.create_table('workouts',
        sa.Column('id', sa.Integer(), server_default=sa.text("nextval('workouts_id_seq'::regclass)"), autoincrement=True, nullable=False),
        sa.Column('uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('sport_id', sa.Integer(), nullable=False),
        sa.Column('title', sa.String(length=255), nullable=True),
        sa.Column('gpx', sa.String(length=255), nullable=True),
        sa.Column('creation_date', sa.DateTime(), nullable=True),
        sa.Column('modification_date', sa.DateTime(), nullable=True),
        sa.Column('workout_date', sa.DateTime(), nullable=False),
        sa.Column('duration', sa.Interval(), nullable=False),
        sa.Column('pauses', sa.Interval(), nullable=True),
        sa.Column('moving', sa.Interval(), nullable=True),
        sa.Column('distance', sa.Numeric(precision=6, scale=3), nullable=True),
        sa.Column('min_alt', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('max_alt', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('descent', sa.Numeric(precision=7, scale=2), nullable=True),
        sa.Column('ascent', sa.Numeric(precision=7, scale=2), nullable=True),
        sa.Column('max_speed', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('ave_speed', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('bounds', postgresql.ARRAY(sa.Float()), nullable=True),
        sa.Column('map', sa.String(length=255), nullable=True),
        sa.Column('map_id', sa.String(length=50), nullable=True),
        sa.Column('weather_start', sa.JSON(), nullable=True),
        sa.Column('weather_end', sa.JSON(), nullable=True),
        sa.Column('notes', sa.String(length=500), nullable=True),
        sa.ForeignKeyConstraint(['sport_id'], ['sports.id'], ),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid')
    )
    op.execute(
        "SELECT setval('workouts_id_seq', (SELECT max(id) FROM activities));"
    )

    op.create_table('workout_segments',
        sa.Column('workout_id', sa.Integer(), nullable=False),
        sa.Column('workout_uuid', postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column('segment_id', sa.Integer(), nullable=False),
        sa.Column('duration', sa.Interval(), nullable=False),
        sa.Column('pauses', sa.Interval(), nullable=True),
        sa.Column('moving', sa.Interval(), nullable=True),
        sa.Column('distance', sa.Numeric(precision=6, scale=3), nullable=True),
        sa.Column('min_alt', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('max_alt', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('descent', sa.Numeric(precision=7, scale=2), nullable=True),
        sa.Column('ascent', sa.Numeric(precision=7, scale=2), nullable=True),
        sa.Column('max_speed', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.Column('ave_speed', sa.Numeric(precision=6, scale=2), nullable=True),
        sa.ForeignKeyConstraint(['workout_id'], ['workouts.id'], ),
        sa.PrimaryKeyConstraint('workout_id', 'segment_id')
    )

    op.add_column('records', sa.Column('workout_date', sa.DateTime(), nullable=True))
    op.add_column('records', sa.Column('workout_id', sa.Integer(), nullable=True))
    op.add_column('records', sa.Column('workout_uuid', postgresql.UUID(as_uuid=True), nullable=True))
    op.create_foreign_key('records_workout_id_fkey', 'records', 'workouts', ['workout_id'], ['id'])

    op.execute(
        """
        INSERT INTO workouts (id, user_id, sport_id, title, gpx, 
        creation_date, modification_date, workout_date, duration, pauses, 
        moving, distance, min_alt, max_alt, descent, ascent, max_speed, 
        ave_speed, bounds, map, map_id, weather_start, weather_end, notes, uuid)
        SELECT id, user_id, sport_id, title, REPLACE(gpx, 'activities/', 'workouts/'), 
        creation_date, modification_date, activity_date, duration, pauses, 
        moving, distance, min_alt, max_alt, descent, ascent, max_speed, 
        ave_speed, bounds, REPLACE(map, 'activities/', 'workouts/'), map_id, 
        weather_start, weather_end, notes, uuid 
        FROM activities;
        """
    )

    op.execute(
        """
        INSERT INTO workout_segments (workout_id, workout_uuid, segment_id,
        duration, pauses, moving, distance, min_alt, max_alt, descent, 
        ascent, max_speed, ave_speed)
         SELECT activity_id, activity_uuid, segment_id,
        duration, pauses, moving, distance, min_alt, max_alt, descent, 
        ascent, max_speed, ave_speed FROM activity_segments;
        """
    )

    op.execute(
        'UPDATE records SET workout_id = activity_id, '
        '                   workout_uuid = activity_uuid, '
        '                   workout_date = activity_date;'
    )

    op.alter_column('records', 'workout_date', nullable=False)
    op.alter_column('records', 'workout_id', nullable=False)
    op.alter_column('records', 'workout_uuid', nullable=False)
    op.drop_column('records', 'activity_date')
    op.drop_column('records', 'activity_id')
    op.drop_column('records', 'activity_uuid')

    op.drop_table('activity_segments')

    op.drop_table('activities')

    rename_upload_folder('activities', 'workouts')


def downgrade():
    op.create_table('activities',
        sa.Column('id', sa.INTEGER(), server_default=sa.text("nextval('activities_id_seq'::regclass)"), autoincrement=True, nullable=False),
        sa.Column('user_id', sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column('sport_id', sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column('title', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column('gpx', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column('creation_date', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
        sa.Column('modification_date', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
        sa.Column('activity_date', postgresql.TIMESTAMP(), autoincrement=False, nullable=False),
        sa.Column('duration', postgresql.INTERVAL(), autoincrement=False, nullable=False),
        sa.Column('pauses', postgresql.INTERVAL(), autoincrement=False, nullable=True),
        sa.Column('moving', postgresql.INTERVAL(), autoincrement=False, nullable=True),
        sa.Column('distance', sa.NUMERIC(precision=6, scale=3), autoincrement=False, nullable=True),
        sa.Column('min_alt', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('max_alt', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('descent', sa.NUMERIC(precision=7, scale=2), autoincrement=False, nullable=True),
        sa.Column('ascent', sa.NUMERIC(precision=7, scale=2), autoincrement=False, nullable=True),
        sa.Column('max_speed', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('ave_speed', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('bounds', postgresql.ARRAY(postgresql.DOUBLE_PRECISION(precision=53)), autoincrement=False, nullable=True),
        sa.Column('map', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column('map_id', sa.VARCHAR(length=50), autoincrement=False, nullable=True),
        sa.Column('weather_end', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
        sa.Column('weather_start', postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
        sa.Column('notes', sa.VARCHAR(length=500), autoincrement=False, nullable=True),
        sa.Column('uuid', postgresql.UUID(), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(['sport_id'], ['sports.id'], name='activities_sport_id_fkey'),
        sa.ForeignKeyConstraint(['user_id'], ['users.id'], name='activities_user_id_fkey'),
        sa.PrimaryKeyConstraint('id', name='activities_pkey'),
        sa.UniqueConstraint('uuid', name='activities_uuid_key'),
        postgresql_ignore_search_path=False
    )
    op.execute(
        "SELECT setval('activities_id_seq', (SELECT max(id) FROM workouts));"
    )

    op.create_table('activity_segments',
        sa.Column('activity_id', sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column('segment_id', sa.INTEGER(), autoincrement=False, nullable=False),
        sa.Column('duration', postgresql.INTERVAL(), autoincrement=False, nullable=False),
        sa.Column('pauses', postgresql.INTERVAL(), autoincrement=False, nullable=True),
        sa.Column('moving', postgresql.INTERVAL(), autoincrement=False, nullable=True),
        sa.Column('distance', sa.NUMERIC(precision=6, scale=3), autoincrement=False, nullable=True),
        sa.Column('min_alt', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('max_alt', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('descent', sa.NUMERIC(precision=7, scale=2), autoincrement=False, nullable=True),
        sa.Column('ascent', sa.NUMERIC(precision=7, scale=2), autoincrement=False, nullable=True),
        sa.Column('max_speed', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('ave_speed', sa.NUMERIC(precision=6, scale=2), autoincrement=False, nullable=True),
        sa.Column('activity_uuid', postgresql.UUID(), autoincrement=False, nullable=False),
        sa.ForeignKeyConstraint(['activity_id'], ['activities.id'], name='activity_segments_activity_id_fkey'),
        sa.PrimaryKeyConstraint('activity_id', 'segment_id', name='activity_segments_pkey')
    )

    op.execute(
        """
        INSERT INTO activities (id, user_id, sport_id, title, gpx, 
        creation_date, modification_date, activity_date, duration, pauses, 
        moving, distance, min_alt, max_alt, descent, ascent, max_speed, 
        ave_speed, bounds, map, map_id, weather_start, weather_end, notes, uuid)
        SELECT id, user_id, sport_id, title, REPLACE(gpx, 'workouts/', 'activities/'),
        creation_date, modification_date, workout_date, duration, pauses, 
        moving, distance, min_alt, max_alt, descent, ascent, max_speed, 
        ave_speed, bounds, REPLACE(map, 'workouts/', 'activities/'), map_id, 
        weather_start, weather_end, notes, uuid 
        FROM workouts;
        """
    )

    op.execute(
        """
        INSERT INTO activity_segments (activity_id, activity_uuid, segment_id,
        duration, pauses, moving, distance, min_alt, max_alt, descent, 
        ascent, max_speed, ave_speed)
         SELECT workout_id, workout_uuid, segment_id,
        duration, pauses, moving, distance, min_alt, max_alt, descent, 
        ascent, max_speed, ave_speed FROM workout_segments;
        """
    )

    op.add_column('records', sa.Column('activity_uuid', postgresql.UUID(), autoincrement=False, nullable=True))
    op.add_column('records', sa.Column('activity_id', sa.INTEGER(), autoincrement=False, nullable=True))
    op.add_column('records', sa.Column('activity_date', postgresql.TIMESTAMP(), autoincrement=False, nullable=True))
    op.create_foreign_key('records_activity_id_fkey', 'records', 'activities', ['activity_id'], ['id'])

    op.execute(
        'UPDATE records SET activity_id = workout_id, '
        '                   activity_uuid = workout_uuid, '
        '                   activity_date = workout_date;'
    )

    op.alter_column('records', 'activity_date', nullable=False)
    op.alter_column('records', 'activity_id', nullable=False)
    op.alter_column('records', 'activity_uuid', nullable=False)
    op.drop_column('records', 'workout_uuid')
    op.drop_column('records', 'workout_id')
    op.drop_column('records', 'workout_date')

    op.drop_table('workout_segments')

    op.drop_table('workouts')

    rename_upload_folder('workouts', 'activities')



"""add static map url to 'Activity' table

Revision ID: 5a42db64e872
Revises: 92adde6ac0d0
Create Date: 2018-05-30 10:52:33.433687

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5a42db64e872'
down_revision = '92adde6ac0d0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('activities', sa.Column('map', sa.String(length=255), nullable=True))
    op.create_unique_constraint(None, 'sports', ['img'])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'sports', type_='unique')
    op.drop_column('activities', 'map')
    # ### end Alembic commands ###



import logging
from typing import Optional

import click
from humanize import naturalsize

from fittrackee import db
from fittrackee.cli.app import app
from fittrackee.users.exceptions import UserNotFoundException
from fittrackee.users.export_data import (
    clean_user_data_export,
    generate_user_data_archives,
)
from fittrackee.users.utils.admin import UserManagerService
from fittrackee.users.utils.token import clean_blacklisted_tokens

handler = logging.StreamHandler()
logger = logging.getLogger('fittrackee_users_cli')
logger.setLevel(logging.INFO)
logger.addHandler(handler)


@click.group(name='users')
def users_cli() -> None:
    """Manage users."""
    pass


@users_cli.command('create')
@click.argument('username')
@click.option('--email', type=str, required=True, help='User email.')
@click.option(
    '--password',
    type=str,
    help='User password. If not provided, a random password is generated.',
)
def create_user(username: str, email: str, password: Optional[str]) -> None:
    """Create an active user account."""
    with app.app_context():
        try:
            user_manager_service = UserManagerService(username)
            user, user_password = user_manager_service.create_user(
                email=email, password=password, check_email=True
            )
            db.session.add(user)
            db.session.commit()
            user_manager_service.update(activate=True)
            click.echo(f"User '{username}' created.")
            if not password:
                click.echo(f"The user password is: {user_password}")
        except Exception as e:
            click.echo(f'Error(s) occurred:\n{e}', err=True)


@users_cli.command('update')
@click.argument('username')
@click.option(
    '--set-admin',
    type=bool,
    help='Add/remove admin rights (when adding admin rights, '
    'it also activates user account if not active).',
)
@click.option('--activate', is_flag=True, help='Activate user account.')
@click.option(
    '--reset-password',
    is_flag=True,
    help='Reset user password (a new password will be displayed).',
)
@click.option('--update-email', type=str, help='Update user email.')
def manage_user(
    username: str,
    set_admin: Optional[bool],
    activate: bool,
    reset_password: bool,
    update_email: Optional[str],
) -> None:
    """Manage given user account."""
    with app.app_context():
        try:
            user_manager_service = UserManagerService(username)
            _, is_user_updated, password = user_manager_service.update(
                is_admin=set_admin,
                with_confirmation=False,
                activate=activate,
                reset_password=reset_password,
                new_email=update_email,
            )
            if is_user_updated:
                click.echo(f"User '{username}' updated.")
                if password:
                    click.echo(f"The new password is: {password}")
            else:
                click.echo("No updates.")
        except UserNotFoundException:
            click.echo(
                f"User '{username}' not found.\n"
                "Check the provided user name (case sensitive).",
                err=True,
            )
        except Exception as e:
            click.echo(f'An error occurred: {e}', err=True)


@users_cli.command('clean_tokens')
@click.option('--days', type=int, required=True, help='Number of days.')
def clean(
    days: int,
) -> None:
    """
    Clean blacklisted tokens expired for more than provided number of days.
    """
    with app.app_context():
        deleted_rows = clean_blacklisted_tokens(days)
        logger.info(f'Blacklisted tokens deleted: {deleted_rows}.')


@users_cli.command('clean_archives')
@click.option('--days', type=int, required=True, help='Number of days.')
def clean_export_archives(
    days: int,
) -> None:
    """
    Clean user export archives created for more than provided number of days.
    """
    with app.app_context():
        counts = clean_user_data_export(days)
        logger.info(
            f'Deleted data export requests: {counts["deleted_requests"]}.'
        )
        logger.info(
            f'Deleted data export archives: {counts["deleted_archives"]}.'
        )
        logger.info(f'Freed space: {naturalsize(counts["freed_space"])}.')


@users_cli.command('export_archives')
@click.option(
    '--max',
    type=int,
    required=True,
    help='Maximum number of export requests to process.',
)
def export_archives(
    max: int,
) -> None:
    """
    Export user data in zip archive if incomplete requests exist.
    To use in case redis is not set.
    """
    with app.app_context():
        count = generate_user_data_archives(max)
        logger.info(f'Generated data export archives: {count}.')



import os
from datetime import datetime
from typing import Any, Dict, Optional, Union

import jwt
from flask import current_app
from sqlalchemy import and_, func
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.expression import select

from fittrackee import appLog, bcrypt, db
from fittrackee.files import get_absolute_file_path
from fittrackee.workouts.models import Workout

from .exceptions import UserNotFoundException
from .roles import UserRole
from .utils.token import decode_user_token, get_user_token

BaseModel: DeclarativeMeta = db.Model


class User(BaseModel):
    __tablename__ = 'users'
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    username = db.Column(db.String(255), unique=True, nullable=False)
    email = db.Column(db.String(255), unique=True, nullable=False)
    password = db.Column(db.String(255), nullable=False)
    created_at = db.Column(db.DateTime, nullable=False)
    admin = db.Column(db.Boolean, default=False, nullable=False)
    first_name = db.Column(db.String(80), nullable=True)
    last_name = db.Column(db.String(80), nullable=True)
    birth_date = db.Column(db.DateTime, nullable=True)
    location = db.Column(db.String(80), nullable=True)
    bio = db.Column(db.String(200), nullable=True)
    picture = db.Column(db.String(255), nullable=True)
    timezone = db.Column(db.String(50), nullable=True)
    date_format = db.Column(db.String(50), nullable=True)
    # does the week start Monday?
    weekm = db.Column(db.Boolean, default=False, nullable=False)
    language = db.Column(db.String(50), nullable=True)
    imperial_units = db.Column(db.Boolean, default=False, nullable=False)
    is_active = db.Column(db.Boolean, default=False, nullable=False)
    email_to_confirm = db.Column(db.String(255), nullable=True)
    confirmation_token = db.Column(db.String(255), nullable=True)
    display_ascent = db.Column(db.Boolean, default=True, nullable=False)
    accepted_policy_date = db.Column(db.DateTime, nullable=True)
    start_elevation_at_zero = db.Column(
        db.Boolean, default=True, nullable=False
    )
    use_raw_gpx_speed = db.Column(db.Boolean, default=False, nullable=False)
    use_dark_mode = db.Column(db.Boolean, default=False, nullable=True)

    workouts = db.relationship(
        'Workout',
        lazy=True,
        backref=db.backref('user', lazy='joined', single_parent=True),
    )
    records = db.relationship(
        'Record',
        lazy=True,
        backref=db.backref('user', lazy='joined', single_parent=True),
    )
    equipments = db.relationship(
        'Equipment',
        lazy='select',
        backref=db.backref('user', lazy='select', single_parent=True),
    )

    def __repr__(self) -> str:
        return f'<User {self.username!r}>'

    def __init__(
        self,
        username: str,
        email: str,
        password: str,
        created_at: Optional[datetime] = None,
    ) -> None:
        self.username = username
        self.email = email
        self.password = bcrypt.generate_password_hash(
            password, current_app.config.get('BCRYPT_LOG_ROUNDS')
        ).decode()
        self.created_at = (
            datetime.utcnow() if created_at is None else created_at
        )

    @staticmethod
    def encode_auth_token(user_id: int) -> str:
        """
        Generates the auth token
        :param user_id: -
        :return: JWToken
        """
        return get_user_token(user_id)

    @staticmethod
    def encode_password_reset_token(user_id: int) -> str:
        """
        Generates the auth token
        :param user_id: -
        :return: JWToken
        """
        return get_user_token(user_id, password_reset=True)

    @staticmethod
    def decode_auth_token(auth_token: str) -> Union[int, str]:
        """
        Decodes the auth token
        :param auth_token: -
        :return: integer|string
        """
        try:
            resp = decode_user_token(auth_token)
            is_blacklisted = BlacklistedToken.check(auth_token)
            if is_blacklisted:
                return 'blacklisted token, please log in again'
            return resp
        except jwt.ExpiredSignatureError:
            return 'signature expired, please log in again'
        except jwt.InvalidTokenError:
            return 'invalid token, please log in again'

    def check_password(self, password: str) -> bool:
        return bcrypt.check_password_hash(self.password, password)

    @staticmethod
    def generate_password_hash(new_password: str) -> str:
        return bcrypt.generate_password_hash(
            new_password, current_app.config.get('BCRYPT_LOG_ROUNDS')
        ).decode()

    def get_user_id(self) -> int:
        return self.id

    @hybrid_property
    def workouts_count(self) -> int:
        return Workout.query.filter(Workout.user_id == self.id).count()

    @workouts_count.expression  # type: ignore
    def workouts_count(self) -> int:
        return (
            select([func.count(Workout.id)])
            .where(Workout.user_id == self.id)
            .label('workouts_count')
        )

    def serialize(self, current_user: 'User') -> Dict:
        role = (
            UserRole.AUTH_USER
            if current_user.id == self.id
            else UserRole.ADMIN
            if current_user.admin
            else UserRole.USER
        )

        if role == UserRole.USER:
            raise UserNotFoundException()

        sports = []
        total = (0, '0:00:00', 0)
        if self.workouts_count > 0:  # type: ignore
            sports = (
                db.session.query(Workout.sport_id)
                .filter(Workout.user_id == self.id)
                .group_by(Workout.sport_id)
                .order_by(Workout.sport_id)
                .all()
            )
            total = (
                db.session.query(
                    func.sum(Workout.distance),
                    func.sum(Workout.duration),
                    func.sum(Workout.ascent),
                )
                .filter(Workout.user_id == self.id)
                .first()
            )

        serialized_user = {
            'admin': self.admin,
            'bio': self.bio,
            'birth_date': self.birth_date,
            'created_at': self.created_at,
            'email': self.email,
            'email_to_confirm': self.email_to_confirm,
            'first_name': self.first_name,
            'is_active': self.is_active,
            'last_name': self.last_name,
            'location': self.location,
            'nb_sports': len(sports),
            'nb_workouts': self.workouts_count,
            'picture': self.picture is not None,
            'records': [record.serialize() for record in self.records],
            'sports_list': [
                sport for sportslist in sports for sport in sportslist
            ],
            'total_ascent': float(total[2]) if total[2] else 0.0,
            'total_distance': float(total[0]),
            'total_duration': str(total[1]),
            'username': self.username,
        }
        if role == UserRole.AUTH_USER:
            accepted_privacy_policy = False
            if self.accepted_policy_date:
                accepted_privacy_policy = (
                    True
                    if current_app.config['privacy_policy_date'] is None
                    else current_app.config['privacy_policy_date']
                    < self.accepted_policy_date
                )
            serialized_user = {
                **serialized_user,
                **{
                    'accepted_privacy_policy': accepted_privacy_policy,
                    'date_format': self.date_format,
                    'display_ascent': self.display_ascent,
                    'imperial_units': self.imperial_units,
                    'language': self.language,
                    'start_elevation_at_zero': self.start_elevation_at_zero,
                    'timezone': self.timezone,
                    'use_dark_mode': self.use_dark_mode,
                    'use_raw_gpx_speed': self.use_raw_gpx_speed,
                    'weekm': self.weekm,
                },
            }

        return serialized_user


UserSportPreferenceEquipment = db.Table(
    'users_sports_preferences_equipments',
    db.Column(
        'user_id',
        db.Integer,
        db.ForeignKey('users.id', ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        'sport_id',
        db.Integer,
        db.ForeignKey('sports.id', ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        'equipment_id',
        db.Integer,
        db.ForeignKey('equipments.id', ondelete="CASCADE"),
        primary_key=True,
    ),
)


class UserSportPreference(BaseModel):
    __tablename__ = 'users_sports_preferences'

    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id'),
        primary_key=True,
    )
    sport_id = db.Column(
        db.Integer,
        db.ForeignKey('sports.id'),
        primary_key=True,
    )
    color = db.Column(db.String(50), nullable=True)
    is_active = db.Column(db.Boolean, default=True, nullable=False)
    stopped_speed_threshold = db.Column(db.Float, default=1.0, nullable=False)

    default_equipments = db.relationship(
        'Equipment',
        secondary=UserSportPreferenceEquipment,
        primaryjoin=and_(
            user_id == UserSportPreferenceEquipment.c.user_id,
            sport_id == UserSportPreferenceEquipment.c.sport_id,
        ),
        lazy='dynamic',
        viewonly=True,
        backref=db.backref('default_for_sports', lazy='select'),
    )

    def __init__(
        self,
        user_id: int,
        sport_id: int,
        stopped_speed_threshold: float,
    ) -> None:
        self.user_id = user_id
        self.sport_id = sport_id
        self.is_active = True
        self.stopped_speed_threshold = stopped_speed_threshold

    def serialize(self) -> Dict:
        return {
            'user_id': self.user_id,
            'sport_id': self.sport_id,
            'color': self.color,
            'is_active': self.is_active,
            'stopped_speed_threshold': self.stopped_speed_threshold,
            'default_equipments': [
                equipment.serialize()
                for equipment in self.default_equipments.all()
            ],
        }


class BlacklistedToken(BaseModel):
    __tablename__ = 'blacklisted_tokens'

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    token = db.Column(db.String(500), unique=True, nullable=False)
    expired_at = db.Column(db.Integer, nullable=False)
    blacklisted_on = db.Column(db.DateTime, nullable=False)

    def __init__(
        self, token: str, blacklisted_on: Optional[datetime] = None
    ) -> None:
        payload = jwt.decode(
            token,
            current_app.config['SECRET_KEY'],
            algorithms=['HS256'],
        )
        self.token = token
        self.expired_at = payload['exp']
        self.blacklisted_on = (
            blacklisted_on if blacklisted_on else datetime.utcnow()
        )

    @classmethod
    def check(cls, auth_token: str) -> bool:
        return cls.query.filter_by(token=str(auth_token)).first() is not None


class UserDataExport(BaseModel):
    __tablename__ = 'users_data_export'

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    user_id = db.Column(
        db.Integer,
        db.ForeignKey('users.id', ondelete='CASCADE'),
        index=True,
        unique=True,
    )
    created_at = db.Column(
        db.DateTime, nullable=False, default=datetime.utcnow
    )
    updated_at = db.Column(
        db.DateTime, nullable=True, onupdate=datetime.utcnow
    )
    completed = db.Column(db.Boolean, nullable=False, default=False)
    file_name = db.Column(db.String(100), nullable=True)
    file_size = db.Column(db.Integer, nullable=True)

    def __init__(
        self,
        user_id: int,
        created_at: Optional[datetime] = None,
    ):
        self.user_id = user_id
        self.created_at = (
            datetime.utcnow() if created_at is None else created_at
        )

    def serialize(self) -> Dict:
        if self.completed:
            status = "successful" if self.file_name else "errored"
        else:
            status = "in_progress"
        return {
            "created_at": self.created_at,
            "status": status,
            "file_name": self.file_name if status == "successful" else None,
            "file_size": self.file_size if status == "successful" else None,
        }


@listens_for(UserDataExport, 'after_delete')
def on_users_data_export_delete(
    mapper: Mapper, connection: Connection, old_record: 'UserDataExport'
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        if old_record.file_name:
            try:
                file_path = (
                    f"exports/{old_record.user_id}/{old_record.file_name}"
                )
                os.remove(get_absolute_file_path(file_path))
            except OSError:
                appLog.error('archive found when deleting export request')



import time
from typing import Any, Dict, Optional

from authlib.integrations.sqla_oauth2 import (
    OAuth2AuthorizationCodeMixin,
    OAuth2ClientMixin,
    OAuth2TokenMixin,
)
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session

from fittrackee import db

BaseModel: DeclarativeMeta = db.Model


class OAuth2Client(BaseModel, OAuth2ClientMixin):
    __tablename__ = 'oauth2_client'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('users.id', ondelete='CASCADE'), index=True
    )
    user = db.relationship('User')

    def serialize(self, with_secret: bool = False) -> Dict:
        client = {
            'client_id': self.client_id,
            'client_description': self.client_description,
            'id': self.id,
            'issued_at': time.strftime(
                '%a, %d %B %Y %H:%M:%S GMT',
                time.gmtime(self.client_id_issued_at),
            ),
            'name': self.client_name,
            'redirect_uris': self.redirect_uris,
            'scope': self.scope,
            'website': self.client_uri,
        }
        if with_secret:
            client['client_secret'] = self.client_secret
        return client

    @property
    def client_description(self) -> Optional[str]:
        return self.client_metadata.get('client_description')


@listens_for(OAuth2Client, 'after_delete')
def on_old_oauth2_delete(
    mapper: Mapper, connection: Connection, old_oauth2_client: OAuth2Client
) -> None:
    @listens_for(db.Session, 'after_flush', once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        session.query(OAuth2AuthorizationCode).filter(
            OAuth2AuthorizationCode.client_id == old_oauth2_client.client_id
        ).delete(synchronize_session=False)
        session.query(OAuth2Token).filter(
            OAuth2Token.client_id == old_oauth2_client.client_id
        ).delete(synchronize_session=False)


class OAuth2AuthorizationCode(BaseModel, OAuth2AuthorizationCodeMixin):
    __tablename__ = 'oauth2_code'
    __table_args__ = (
        db.Index(
            'ix_oauth2_code_client_id',
            'client_id',
        ),
    )

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('users.id', ondelete='CASCADE'), index=True
    )
    user = db.relationship('User')


class OAuth2Token(BaseModel, OAuth2TokenMixin):
    __tablename__ = 'oauth2_token'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey('users.id', ondelete='CASCADE'), index=True
    )
    user = db.relationship('User')

    def is_refresh_token_active(self) -> bool:
        if self.is_revoked():
            return False
        expires_at = self.issued_at + self.expires_in * 2
        return expires_at >= time.time()

    @classmethod
    def revoke_client_tokens(cls, client_id: str) -> None:
        sql = """
            UPDATE oauth2_token
            SET access_token_revoked_at = %(revoked_at)s
            WHERE client_id = %(client_id)s;
        """
        db.engine.execute(
            sql, {'client_id': client_id, 'revoked_at': int(time.time())}
        )
        db.session.commit()



import click

from fittrackee.migrations.commands import db_cli
from fittrackee.oauth2.commands import oauth2_cli
from fittrackee.users.commands import users_cli


@click.group()
def cli() -> None:
    """FitTrackee Command Line Interface"""
    pass


cli.add_command(db_cli)
cli.add_command(oauth2_cli)
cli.add_command(users_cli)



"""replace 'is_default' with 'is_active' in 'Sports' table

Revision ID: 1345afe3b11d
Revises: f69f1e413bde
Create Date: 2019-09-22 17:57:00.595775

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1345afe3b11d'
down_revision = 'f69f1e413bde'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('sports',
                  sa.Column('is_active',
                            sa.Boolean(create_constraint=50),
                            nullable=True))
    op.execute("UPDATE sports SET is_active = true")
    op.alter_column('sports', 'is_active', nullable=False)
    op.drop_column('sports', 'is_default')


def downgrade():
    op.add_column('sports',
                  sa.Column('is_default',
                            sa.Boolean(create_constraint=50),
                            nullable=True))
    op.execute("UPDATE sports SET is_default = true")
    op.alter_column('sports', 'is_default', nullable=False)
    op.drop_column('sports', 'is_active')



"""add 'weekm' in 'User' table

Revision ID: 27425324c9e3
Revises: 096dd0b43beb
Create Date: 2019-08-31 15:25:09.412426

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '27425324c9e3'
down_revision = '096dd0b43beb'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('users', sa.Column('weekm', sa.Boolean(create_constraint=50), nullable=True))
    op.execute("UPDATE users SET weekm = false")
    op.alter_column('users', 'weekm', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'weekm')
    # ### end Alembic commands ###



from time import time
from typing import Dict

from werkzeug.security import gen_salt

from fittrackee.users.models import User

from .exceptions import InvalidOAuth2Scopes
from .models import OAuth2Client

VALID_SCOPES = [
    'application:write',
    'equipments:read',
    'equipments:write',
    'profile:read',
    'profile:write',
    'users:read',
    'users:write',
    'workouts:read',
    'workouts:write',
]


def check_scope(scope: str) -> str:
    """
    Verify if provided scope is valid.
    """
    if not isinstance(scope, str) or not scope:
        raise InvalidOAuth2Scopes()

    valid_scopes = []
    scopes = scope.split()
    for value in scopes:
        if value in VALID_SCOPES:
            valid_scopes.append(value)

    if not valid_scopes:
        raise InvalidOAuth2Scopes()

    return ' '.join(valid_scopes)


def create_oauth2_client(metadata: Dict, user: User) -> OAuth2Client:
    """
    Create OAuth2 client for 3rd-party applications.

    Only Authorization Code Grant with 'client_secret_post' as method
    is supported.
    Code challenge can be used if provided on authorization.
    """
    client_metadata = {
        'client_name': metadata['client_name'],
        'client_description': metadata.get('client_description'),
        'client_uri': metadata['client_uri'],
        'redirect_uris': metadata['redirect_uris'],
        'scope': check_scope(metadata['scope']),
        'grant_types': ['authorization_code', 'refresh_token'],
        'response_types': ['code'],
        'token_endpoint_auth_method': 'client_secret_post',
    }
    client_id = gen_salt(24)
    client_id_issued_at = int(time())
    client = OAuth2Client(
        client_id=client_id,
        client_id_issued_at=client_id_issued_at,
        user_id=user.id,
    )
    client.set_client_metadata(client_metadata)
    client.client_secret = gen_salt(48)
    return client



import os
from typing import Type, Union

from dramatiq.brokers.redis import RedisBroker
from dramatiq.brokers.stub import StubBroker
from flask import current_app
from sqlalchemy.pool import NullPool

broker: Union[Type['RedisBroker'], Type['StubBroker']] = (
    StubBroker
    if os.getenv("APP_SETTINGS") == "fittrackee.config.TestingConfig"
    else RedisBroker
)

XDIST_WORKER = (
    f"_{os.getenv('PYTEST_XDIST_WORKER')}"
    if os.getenv('PYTEST_XDIST_WORKER')
    else ''
)


class BaseConfig:
    DEBUG = False
    TESTING = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    BCRYPT_LOG_ROUNDS = 13
    TOKEN_EXPIRATION_DAYS = 30
    TOKEN_EXPIRATION_SECONDS = 0
    PASSWORD_TOKEN_EXPIRATION_SECONDS = 3600
    UPLOAD_FOLDER = os.path.join(
        os.getenv('UPLOAD_FOLDER', current_app.root_path), 'uploads'
    )
    PICTURE_ALLOWED_EXTENSIONS = {'jpg', 'png', 'gif'}
    WORKOUT_ALLOWED_EXTENSIONS = {'gpx', 'zip'}
    TEMPLATES_FOLDER = os.path.join(current_app.root_path, 'emails/templates')
    UI_URL = os.environ.get('UI_URL')
    EMAIL_URL = os.environ.get('EMAIL_URL')
    SENDER_EMAIL = os.environ.get('SENDER_EMAIL')
    CAN_SEND_EMAILS = False
    DRAMATIQ_BROKER = broker
    TILE_SERVER = {
        'URL': os.environ.get(
            'TILE_SERVER_URL',
            'https://tile.openstreetmap.org/{z}/{x}/{y}.png',
        ),
        'ATTRIBUTION': os.environ.get(
            'MAP_ATTRIBUTION',
            '&copy; <a href="http://www.openstreetmap.org/copyright" '
            'target="_blank" rel="noopener noreferrer">OpenStreetMap</a>'
            ' contributors',
        ),
        'DEFAULT_STATICMAP': (
            os.environ.get('DEFAULT_STATICMAP', 'false').lower() == 'true'
        ),
        'STATICMAP_SUBDOMAINS': os.environ.get('STATICMAP_SUBDOMAINS', ''),
    }
    TRANSLATIONS_FOLDER = os.path.join(
        current_app.root_path, 'emails/translations'
    )
    LANGUAGES = [
        'en',
        'fr',
        'de',
        'it',
        'nb',
        'nl',
        'es',
        'gl',
        'pl',
        'eu',
        'cs',
    ]
    OAUTH2_TOKEN_EXPIRES_IN = {
        'authorization_code': 864000,  # 10 days
        'refresh_token': 864000,  # 10 days
    }
    OAUTH2_REFRESH_TOKEN_GENERATOR = True
    DATA_EXPORT_EXPIRATION = 24  # hours


class DevelopmentConfig(BaseConfig):
    DEBUG = True
    SQLALCHEMY_DATABASE_URI = os.environ.get('DATABASE_URL')
    SECRET_KEY = 'development key'  # nosec
    BCRYPT_LOG_ROUNDS = 4
    DRAMATIQ_BROKER_URL = os.getenv('REDIS_URL', 'redis://')


class TestingConfig(BaseConfig):
    DEBUG = True
    TESTING = True
    SQLALCHEMY_DATABASE_URI = (
        os.environ.get('DATABASE_TEST_URL', '') + XDIST_WORKER
    )
    UPLOAD_FOLDER = os.path.join(
        os.getenv('UPLOAD_FOLDER', current_app.root_path),
        'uploads' + XDIST_WORKER,
    )
    SECRET_KEY = 'test key'  # nosec
    BCRYPT_LOG_ROUNDS = 4
    TOKEN_EXPIRATION_DAYS = 0
    TOKEN_EXPIRATION_SECONDS = 60
    PASSWORD_TOKEN_EXPIRATION_SECONDS = 60
    UI_URL = 'http://0.0.0.0:5000'
    SENDER_EMAIL = 'fittrackee@example.com'
    OAUTH2_TOKEN_EXPIRES_IN = {
        'authorization_code': 60,
        'refresh_token': 60,
    }


class End2EndTestingConfig(TestingConfig):
    DRAMATIQ_BROKER_URL = os.getenv('REDIS_URL', 'redis://')


class ProductionConfig(BaseConfig):
    DEBUG = False
    # https://docs.sqlalchemy.org/en/13/core/pooling.html#using-connection-pools-with-multiprocessing-or-os-fork  # noqa
    SQLALCHEMY_ENGINE_OPTIONS = (
        {'poolclass': NullPool}
        if os.getenv('DATABASE_DISABLE_POOLING', 'False') == 'True'
        else {}
    )
    SQLALCHEMY_DATABASE_URI = os.environ.get('DATABASE_URL')
    SECRET_KEY = os.getenv('APP_SECRET_KEY')
    DRAMATIQ_BROKER_URL = os.getenv('REDIS_URL', 'redis://')



from typing import Optional

from flask import current_app


def get_language(language: Optional[str]) -> str:
    # Note: some users may not have language preferences set
    if not language or language not in current_app.config['LANGUAGES']:
        language = 'en'
    return language



from typing import Dict

from fittrackee import dramatiq, email_service


@dramatiq.actor(queue_name='fittrackee_emails')
def reset_password_email(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='password_reset_request',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )


@dramatiq.actor(queue_name='fittrackee_emails')
def email_updated_to_current_address(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='email_update_to_current_email',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )


@dramatiq.actor(queue_name='fittrackee_emails')
def email_updated_to_new_address(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='email_update_to_new_email',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )


@dramatiq.actor(queue_name='fittrackee_emails')
def password_change_email(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='password_change',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )


@dramatiq.actor(queue_name='fittrackee_emails')
def account_confirmation_email(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='account_confirmation',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )


@dramatiq.actor(queue_name='fittrackee_emails')
def data_export_email(user: Dict, email_data: Dict) -> None:
    email_service.send(
        template='data_export_ready',
        lang=user['language'],
        recipient=user['email'],
        data=email_data,
    )



import hashlib
import random
from typing import Dict, List

from flask import current_app
from staticmap import Line, StaticMap

from fittrackee import VERSION
from fittrackee.files import get_absolute_file_path


def get_static_map_tile_server_url(tile_server_config: Dict) -> str:
    if tile_server_config['STATICMAP_SUBDOMAINS']:
        subdomains = tile_server_config['STATICMAP_SUBDOMAINS'].split(',')
        subdomain = f'{random.choice(subdomains)}.'  # nosec
    else:
        subdomain = ''
    return tile_server_config['URL'].replace('{s}.', subdomain)


def generate_map(map_filepath: str, map_data: List) -> None:
    """
    Generate and save map image from map data
    """
    m = StaticMap(400, 225, 10)
    m.headers = {'User-Agent': f'FitTrackee v{VERSION}'}
    if not current_app.config['TILE_SERVER']['DEFAULT_STATICMAP']:
        m.url_template = get_static_map_tile_server_url(
            current_app.config['TILE_SERVER']
        )
    line = Line(map_data, '#3388FF', 4)
    m.add_line(line)
    image = m.render()
    image.save(map_filepath)


def get_map_hash(map_filepath: str) -> str:
    """
    Generate a md5 hash used as id instead of workout id, to retrieve map
    image (maps are sensitive data)
    """
    md5 = hashlib.md5()  # nosec  # need 3.9+ to use 'usedforsecurity' flag
    absolute_map_filepath = get_absolute_file_path(map_filepath)
    with open(absolute_map_filepath, 'rb') as f:
        for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
            md5.update(chunk)
    return md5.hexdigest()



from functools import wraps
from typing import Any, Callable, List, Union

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.oauth2 import OAuth2Error
from authlib.oauth2.rfc6749.errors import MissingAuthorizationError
from flask import current_app, request
from werkzeug.exceptions import RequestEntityTooLarge

from fittrackee.responses import (
    ForbiddenErrorResponse,
    PayloadTooLargeErrorResponse,
    UnauthorizedErrorResponse,
)
from fittrackee.users.models import User


class CustomResourceProtector(ResourceProtector):
    def __call__(
        self,
        scopes: Union[str, List, None] = None,
        as_admin: bool = False,
    ) -> Callable:
        def wrapper(f: Callable) -> Callable:
            @wraps(f)
            def decorated(*args: Any, **kwargs: Any) -> Callable:
                auth_user = None
                auth_header = request.headers.get('Authorization')
                if not auth_header:
                    return UnauthorizedErrorResponse(
                        'provide a valid auth token'
                    )

                # First-party application (Fittrackee front-end)
                # in this case, scopes will be ignored
                auth_token = auth_header.split(' ')[1]
                resp = User.decode_auth_token(auth_token)
                if isinstance(resp, int):
                    auth_user = User.query.filter_by(id=resp).first()

                # Third-party applications
                if not auth_user:
                    current_token = None
                    try:
                        current_token = self.acquire_token(scopes)
                    except MissingAuthorizationError as error:
                        self.raise_error_response(error)
                    except OAuth2Error as error:
                        self.raise_error_response(error)
                    except RequestEntityTooLarge:
                        file_type = ''
                        if request.endpoint in [
                            'auth.edit_picture',
                            'workouts.post_workout',
                        ]:
                            file_type = (
                                'picture'
                                if request.endpoint == 'auth.edit_picture'
                                else 'workout'
                            )
                        return PayloadTooLargeErrorResponse(
                            file_type=file_type,
                            file_size=request.content_length,
                            max_size=current_app.config['MAX_CONTENT_LENGTH'],
                        )
                    auth_user = (
                        None if current_token is None else current_token.user
                    )

                if not auth_user or not auth_user.is_active:
                    return UnauthorizedErrorResponse(
                        'provide a valid auth token'
                    )
                if as_admin and not auth_user.admin:
                    return ForbiddenErrorResponse()
                return f(auth_user, *args, **kwargs)

            return decorated

        return wrapper
